# -*- coding: utf-8 -*-
import operator
import itertools

def main():
    a = tensor(0, [
        tensor(1, [1, 0, 0, 5]),
        tensor(1, [0, 1, 0, 2]),
        tensor(1, [0, 0, 1, 2]),
        tensor(1, [0, 0, 0, 1]),
    ])
    b = tensor(0, [
        tensor(1, [1, 2]),
        tensor(1, [3, 4]),
        tensor(1, [0, 2]),
        tensor(1, [1, 0]),
    ])
    c = matmul(a, b)
    print a
    print b
    print c

def eye(dimensions):
    return generate_tensor(dimensions, eye_gen)
    
def eye_gen(shape):
    if all(count == shape[0] for count in shape):
        return 1
    else:
        return 0

def matmul(a, b):
    return l(a, {1:cnt("k")}) * l(b, {0:"k"})

class Tensor(object):
    def __init__(self, shape, values, indices):
        self.shape = shape
        self.values = values
        self.indices = indices

    def __pos__(self):
        return self

    def __neg__(self):
        return unary_operation(operator.neg, self)

    def __add__(self, other):
        return binary_operation(operator.add, self, other)

    def __radd__(self, other):
        return binary_operation(operator.add, other, self)

    def __sub__(self, other):
        return binary_operation(operator.sub, self, other)

    def __rsub__(self, other):
        return binary_operation(operator.sub, other, self)

    def __mul__(self, other):
        return binary_operation(operator.mul, self, other)

    def __rmul__(self, other):
        return binary_operation(operator.mul, other, self)

    def __repr__(self):
        shape = self.shape
        signature = None
        if all(isinstance(i, int) and i >= 0 for i in shape):
            if len(shape)/2 <= max(shape):
                shape = range(max(shape)+1)
                signature = ""
        if signature is None:
            signature = "{" + ", ".join(repr(i) for i in self.shape) + "}"
        indexer = tensor_reshape(self.shape, self.indices, shape, self.indices)
        return "".join(generate_table(indexer, self.values)) + signature

def generate_table(indexer, values):
    odds = indexer[1::2]
    even = indexer[0::2]
    rows = []
    stride = 1
    for count, _ in odds:
        stride *= count
        rows.append(stride)
    row_stride = stride
    cols = []
    for count, _ in even:
        stride *= count
        cols.append(stride)
    col_stride = stride
    if len(rows) < len(cols):
        rows.append(row_stride)
    if len(rows) > len(cols):
        cols.append(col_stride)

    col_size = [1 for col in range(row_stride)]
    for j, index in enumerate(take_indices(odds + even)):
        k = j % len(col_size)
        col_size[k] = max(col_size[k], len(repr(values[index])))

    view1 = [u"\u256D", u"\u2502", u"\u2570", u"\u256E", u"\u2502", u"\u256F"]
    view2 = [u"\u239B", u"\u239C", u"\u239D", u"\u239E", u"\u239F", u"\u23A0"]
    view = view1

    sp = ""
    for j, i in enumerate(take_indices(odds + even)):
        k = j % len(col_size)
        yield sp
        for z, n in enumerate(reversed(rows)):
            if j % n == 0:
                p = (j - j % row_stride)
                uc = (p % cols[-1-z] == 0)
                lc = ((p + row_stride) % cols[-1-z] == 0)
                if uc and lc:
                    yield u"\u27EE".encode('utf-8')
                elif uc:
                    yield view[0].encode('utf-8')
                elif lc:
                    yield view[2].encode('utf-8')
                else:
                    yield view[1].encode('utf-8')
        yield repr(values[i]).rjust(col_size[k])
        sp = " "
        for z, n in enumerate(rows):
            if (j+1) % n == 0:
                p = (j - j % row_stride)
                uc = ((p + row_stride) % cols[z] == 0)
                lc = (p % cols[z] == 0)
                if uc and lc:
                    yield u"\u27EF".encode('utf-8')
                elif uc:
                    yield view[5].encode('utf-8')
                elif lc:
                    yield view[3].encode('utf-8')
                else:
                    yield view[4].encode('utf-8')
        if (j+1) % len(col_size) == 0:
            sp = "\n"

def unary_operation(op, self):
    return Tensor(self.shape,
        [op(value) for value in self.values], self.indices)

def binary_operation(op, self, other):
    self = to_tensor(self)
    other = to_tensor(other)
    contracts = set()
    shape = []
    indices = {}
    for key in self.shape:
        if key in indices:
            continue
        count1 = self.indices.get(key, 1)
        count2 = other.indices.get(key, 1)
        assert count1 == count2 or count1 == 1 or count2 == 1, "shape error"
        indices[key] = max(count1, count2)
        shape.append(key)
    for key2 in other.shape:
        if key2 in indices:
            continue
        if is_contracting(key2) and contracting_pair(key2) in indices:
            key1 = contracting_pair(key2)
            contracts.add(key1)
        else:
            key1 = key2
            shape.append(key1)
        count1 = self.indices.get(key1, 1)
        count2 = other.indices.get(key2, 1)
        assert count1 == count2 or count1 == 1 or count2 == 1, "shape error"
        indices[key1] = max(count1, count2)
    indexer1 = tensor_reshape(self.shape, self.indices, shape, indices)
    indexer2 = tensor_reshape(other.shape, other.indices, shape, indices)
    indices1 = take_indices(indexer1, 0)
    indices2 = take_indices(indexer2, 0)
    values = [op(self.values[i], other.values[j])
        for i, j in itertools.izip(indices1, indices2)]
    num = Tensor(shape, values, indices)
    for key in contracts:
        num = fold_dimension(contracting_fold(key), num, key)
    return num

def fold_dimension(op, self, rank=0):
    if rank not in self.indices:
        return self
    strides = tensor_strides(self.shape, self.indices)
    count, stride = strides.pop(rank)
    width = count*stride
    shape = [i for i in self.shape if i != rank]
    indices = dict((i, self.indices[i]) for i in shape)
    indexer = [strides[i] for i in shape]
    values = [reduce(op,
        (self.values[j] for j in range(i, i+width, stride)))
        for i in take_indices(indexer, 0)]
    if len(shape) == 0:
        return values[0]
    return Tensor(shape, values, indices)

def index_dimension(self, rank, index):
    if rank not in self.indices:
        return self
    strides = tensor_strides(self.shape, self.indices)
    count, stride = strides.pop(rank)
    assert 0 <= index < count, "index range error"
    width = count*stride
    shape = [i for i in self.shape if i != rank]
    indices = dict((i, self.indices[i]) for i in shape)
    indexer = [strides[i] for i in shape]
    values = [self.values[i] for i in take_indices(indexer, index*stride)]
    if len(shape) == 0:
        return values[0]
    return Tensor(shape, values, indices)

def l(self, new_names):
    strides = tensor_strides(self.shape, self.indices)
    contracts = set()
    indices = {}
    new_strides = {}
    shape = []
    for rank, (count1, stride1) in strides.iteritems():
        rank = new_names.get(rank, rank)
        if is_contracting(rank) and contracting_pair(rank) in indices:
            key = contracting_pair(rank)
            contracts.add(rank)
        if rank in new_strides:
            count2, stride2 = new_strides[rank]
            assert count1 == count2, "shape error"
            new_strides[rank] = count1, stride1+stride2
        else:
            new_strides[rank] = count1, stride1
            indices[rank] = count1
            shape.append(rank)
    indexer = []
    for rank in shape:
        indexer.append(new_strides[rank])
    values = [self.values[i] for i in take_indices(indexer, 0)]
    num = Tensor(shape, values, indices)
    for key in contracts:
        num = fold_dimension(contracting_fold(key), num, key)
    return num

def reverse_dimension(self, rank):
    if rank not in self.indices:
        return self
    offset = 0
    strides = tensor_strides(self.shape, self.indices)
    indices = {}
    shuffles = []
    for index in self.shape:
        count, stride = strides[index]
        if index != rank:
            shuffle = (count, stride, stride-stride*count, count-1)
        else:
            offset += stride*(count-1)
            shuffle = (count, -stride, -stride+stride*count, count-1)
        indices[index] = count
        shuffles.append(shuffle)
    values = [self.values[i] for i in take_shuffle(shuffles, offset)]
    return Tensor(self.shape, values, indices)

def roll_dimension(self, rank, amount):
    if rank not in self.indices:
        return self
    offset = 0
    strides = tensor_strides(self.shape, self.indices)
    indices = {}
    shuffles = []
    for index in self.shape:
        if index != rank:
            shuffle = (count, stride, stride-stride*count, count-1)
        else:
            amount %= count
            offset += stride*amount
            shuffle = (count, stride, stride-stride*count, count-1-amount)
        indices[index] = count
        shuffles.append(shuffle)
    values = [self.values[i] for i in take_shuffle(shuffles, offset)]
    return Tensor(self.shape, values, indices)

def slice_dimension(self, rank, start, stop):
    if rank not in self.indices:
        return self
    offset = 0
    strides = tensor_strides(self.shape, self.indices)
    indices = {}
    shuffles = []
    for index in self.shape:
        if index != rank:
            shuffle = (count, stride, stride-stride*count, count-1)
        else:
            assert 0 <= start < stop <= count, "slice error"
            offset += stride*start
            count = stop-start
            shuffle = (count, stride, stride-stride*count, count-1)
        indices[index] = count
        shuffles.append(shuffle)
    values = [self.values[i] for i in take_shuffle(shuffles, offset)]
    return Tensor(self.shape, values, indices)

def generate_tensor(dimensions, function):
    shape = []
    counts = []
    indices = {}
    for key, count in dimensions:
        assert key not in indices, "shape error"
        indices[key] = count
        counts.append(count)
        shape.append(key)
    values = [function(state) for state in index_generator(counts)]
    return Tensor(shape, values, indices)
    
def index_generator(counts):
    state = [0] * len(counts)
    while True:
        yield state
        for i in range(len(counts)):
            state[i] += 1
            if state[i] < counts[i]:
                break
            else:
                state[i] = 0
        else:
            break

def tensor(index, objects):
    objects = [to_tensor(n) for n in objects]
    shape   = []
    indices = {}
    for t in objects:
        for col, count in t.indices.iteritems():
            assert not is_contracting_pair(index, col), "cannot re-introduce a rank"
            if col not in indices:
                shape.append(col)
                indices[col] = count
            assert indices[col] == count or count == 1, "shape error"
            if is_contracting(col):
                assert contracting_pair(col) not in indices, "polarity error"
            indices[col] = max(indices[col], count)
    values = [t.values[i]
        for t in objects
        for i in take_indices(tensor_reshape(
            t.shape, t.indices, shape, indices))]
    indices[index] = len(objects)
    shape.append(index)
    return Tensor(shape, values, indices)

def tensor_reshape(shape, indices, newshape, newindices):
    vector = tensor_strides(shape, indices)
    new_vector = []
    for index in newshape:
        if is_contracting(index):
            coindex = contracting_pair(index)
            if coindex in vector:
                count1, stride = vector.get(coindex, (1,0))
            else:
                count1, stride = vector.get(index, (1,0))
        else:
            count1, stride = vector.get(index, (1,0))
        count2 = newindices[index]
        assert count1 == count2 or count1 == 1, (index, count1, count2)
        new_vector.append((count2, stride))
    return new_vector
    
def tensor_strides(shape, indices, stride=1):
    vector = {}
    for index in shape:
        count = indices.get(index, 1)
        vector[index] = (count, (stride if count > 1 else 0))
        stride *= count
    return vector

def take_indices(vector, offset=0):
    state = [0] * len(vector)
    while True:
        yield offset
        for i in range(len(vector)):
            offset   += vector[i][1]
            state[i] += 1
            if state[i] < vector[i][0]:
                break
            else:
                offset  -= state[i]*vector[i][1]
                state[i] = 0
        else:
            break

def take_shuffle(shuffles, offset=0):
    state = [0] * len(shuffles)
    while True:
        yield offset
        for i in range(len(shuffles)):
            count, step, skip, skipi = shuffles[i]
            offset += (step if state[i] != skipi else skip)
            state[i] += 1
            if state[i] < count:
                break
            else:
                state[i] = 0
        else:
            break

def is_contracting(index):
    return True

def is_contracting_pair(a, b):
    if isinstance(a, cnt) and a.index == b:
        return True
    if isinstance(b, cnt) and b.index == a:
        return True
    return False

def contracting_pair(index):
    if isinstance(index, cnt):
        return index.index
    else:
        return cnt(index)

def contracting_fold(index):
    if isinstance(index, csym):
        return index.op
    elif isinstance(index, cnt):
        return contracting_fold(index.index)
    else:
        return operator.add

class cnt(object):
    def __init__(self, index):
        self.index = index

    def __eq__(self, other):
        if isinstance(other, cnt):
            return self.index == other.index
        return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.index)

    def __repr__(self):
        return "cnt({})".format(self.index)

class csym(object):
    def __init__(self, op):
        self.op = op

    def __repr__(self):
        return "<csym({}) {}>".format(self.op, id(self))

def to_tensor(value):
    if isinstance(value, (int, float, bool)):
        return Tensor([], [value], {})
    assert isinstance(value, Tensor), "requires support for this value"
    return value

if __name__=="__main__":
    main()
